import os
import time
import ast
from collections import deque
import multiprocessing as mp

import torch
import numpy as np

class BasicDisplay:
    def __init__(self):
        self.losses = deque(maxlen=100)

    def handle(self, data):
        if data["type"] == "train_info":
            epoch = data["epoch"]
            batch_idx = data["batch_idx"]

            self.losses.append(data["loss"])
            loss_val = round(np.mean(self.losses), 2)

            print(f"({epoch:03} {batch_idx:04}) {loss_val:.2f}", end="\r")

        elif data["type"] == "val_info":
            epoch = data["epoch"]
            batch_num = data["batch_idx"]

            val_str = f"Epoch: {epoch} Batch: {batch_num} "

            for loss_info in data["losses"]:
                loss_type = loss_info["loss_type"]
                loss_val = loss_info["loss"]
                val_str += f"{loss_type}: {loss_val} "
            print()
            print(val_str)


class LogWriter:
    """
    Class used for the main process to coordinate writing to log file/saving
    checkpoints.
    """

    def __init__(self, log_path, log_proc=True, display=BasicDisplay):
        os.makedirs(log_path, exist_ok=True)
        self.log_path = log_path
        self.log_file_path = f"{log_path}/train.log"
        self.log_file = open(self.log_file_path, "w")
        
        if log_proc:
            spawn_logger_worker(display, self.log_file_path)

    def log_info(self, info):
        info.dump(self.log_file)
        self.log_file.flush()

    def log_str(self, log_str):
        self.log_file.write(log_str + "\n")
        self.log_file.flush()

    def checkpoint(self, epoch, batch_idx, model):
        checkpoint_path = f"{self.log_path}/model_{epoch}.pth"
        torch.save(model.state_dict(), checkpoint_path)



def log_worker(display_type, log_path):

    logger = display_type()

    file_obj = open(log_path, "r")

    # Main Loop
    while True:
        new_data = file_obj.readline().rstrip()

        if new_data == "":
            time.sleep(0.1)
        else:
            try:
                new_data = ast.literal_eval(new_data)
            except:
                print(f"Failed to parse: {new_data}")
            else:
                logger.handle(new_data)


def spawn_logger_worker(display_type, log_path):
    p = mp.Process(target=log_worker, args=(display_type, log_path,))
    p.daemon = True
    p.start()




import ast
from dataclasses import dataclass
from typing import Dict, List
from mllg.info import TrainInfo, ValidationInfo

@dataclass(frozen=True)
class RunSummary:
    config: Dict
    train_steps: List[TrainInfo]
    val_steps: List[ValidationInfo]

    @classmethod
    def from_file(cls, file_path):
        with open(file_path, 'r') as file_data:
            lines = file_data.readlines()
        
        train_steps = []
        val_steps = []
        for l in lines:
            try:
                data = ast.literal_eval(l)
            except SyntaxError:
                pass

            if data['type'] == 'config':
                config = data
            elif data['type'] == 'train_info':
                train_steps.append(TrainInfo.from_dict(data))
            elif data['type'] == 'val_info':
                val_steps.append(ValidationInfo.from_dict(data))

        return RunSummary(config, train_steps, val_steps)
    
    def train_loss_vals(self):
        loss_vals = [t.loss for t in self.train_steps]
        return loss_vals

    def train_loss_batches(self, batch_size):
        batch_nums = [batch_size * (i + 1) for i in range(len(self.train_steps))]
        return batch_nums

    def validation_vals(self, val_name):
        vals = []
        for val in self.val_steps:
            for loss in val.losses:
                print(loss)
                if loss.loss_type == val_name:
                    vals.append(loss.loss)
        return vals




import time
from typing import Any
from dataclasses import dataclass, asdict


@dataclass(frozen=True)
class TestInfo:
    loss_type: str
    loss: float
    time: float = 0

    def __post_init__(self):
        
        # TODO: This is very bad
        if self.time == 0:
            super().__setattr__("time", time.time())
    
    @classmethod
    def from_dict(cls, data):
        return cls(loss_type=data['loss_type'], loss=data['loss'], time=data['time'])

    def __str__(self):
        dict_val = asdict(self)
        dict_val["type"] = "test_info"
        return str(dict_val)

    def dump(self, file_obj):
        file_obj.write(str(self) + "\n")


@dataclass(frozen=True)
class LossInfo:
    loss_type: str
    loss: float
    time: float = 0

    def __post_init__(self):
        
        # TODO: This is very bad
        if self.time == 0:
            super().__setattr__("time", time.time())
    
    @classmethod
    def from_dict(cls, data):
        return cls(loss_type=data['loss_type'], loss=data['loss'], time=data['time'])

    def __str__(self):
        dict_val = asdict(self)
        dict_val["type"] = "loss_info"
        return str(dict_val)

    def dump(self, file_obj):
        file_obj.write(str(self) + "\n")


@dataclass(frozen=True)
class ValidationInfo:
    epoch: int
    batch_idx: int
    losses: Any
    time: float = 0

    def __post_init__(self):
        if self.time == 0:
            super().__setattr__("time", time.time())

    @classmethod
    def from_dict(cls, data):
        losses = [TestInfo.from_dict(l) for l in data['losses']]
        return cls(epoch=data['epoch'], batch_idx=data['batch_idx'],  losses=losses, time=data['time'])

    def __str__(self):
        dict_val = asdict(self)
        dict_val["type"] = "val_info"
        return str(dict_val)

    def dump(self, file_obj):
        file_obj.write(str(self) + "\n")


@dataclass(frozen=True)
class TrainInfo:
    epoch: int
    batch_idx: int
    loss: float
    time: float = 0

    def __post_init__(self):
        if self.time == 0:
            super().__setattr__("time", time.time())

    @classmethod
    def from_dict(cls, data):
        return cls(epoch=data['epoch'], batch_idx=data['batch_idx'], loss=data['loss'], time=data['time'])

    def __str__(self):
        dict_val = asdict(self)
        dict_val["type"] = "train_info"
        return str(dict_val)

    def dump(self, file_obj):
        file_obj.write(str(self) + "\n")

@dataclass(frozen=True)
class TrainStepInfo:
    epoch: int
    batch_idx: int
    losses: Any
    time: float = 0

    def __post_init__(self):
        if self.time == 0:
            super().__setattr__("time", time.time())

    @classmethod
    def from_dict(cls, data):
        losses = [LossInfo.from_dict(l) for l in data['losses']]
        return cls(epoch=data['epoch'], batch_idx=data['batch_idx'],  losses=losses, time=data['time'])

    def __str__(self):
        dict_val = asdict(self)
        dict_val["type"] = "trainstep_info"
        return str(dict_val)

    def dump(self, file_obj):
        file_obj.write(str(self) + "\n")
